/////////////////
// OS Includes
#include <stdio.h>
#include <stdlib.h>
#include <fstream.h>
#include <iostream.h>
#include <string.h>

////////////////
// Includes from standard C libraries...
#include <sys/times.h>
#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>

//////////////
// Includes
#include "ClauseList.h"
#include "Clause.h"
#include "Random.h"
#include "SATSolver.h"
#include "SATInstance.h"

#include "LoadCnf.h"

/////////////
// Globals

ofstream * PermanentFS;
ofstream * TemporaryFS;
bool RelevanceFlag = TRUE;
bool DumpFlag = FALSE;

/////////////
// Defines

main(int iArgc_, char* aArgv_[])
{
    if (iArgc_<2) {
	cout << "Use: rel_sat_r FILE.cnf {-seed N} {-maxsec N} {-maxbt N}" << endl <<
	    "    {-trials N} {-learnorder N} {-numsol N} {-nosol}" << endl <<
	    "    {-fudge N A B} {-norelevance} {-dump}" << endl;
	exit(1);
    }

    long lClkTck = CLK_TCK;
    unsigned long lSeed;
    struct timeval tv;
    struct timezone tzp;
    gettimeofday(&tv,&tzp);
    lSeed = (( tv.tv_sec & 0177 ) * 1000000) + tv.tv_usec;

    int iNumTrials = 10;
    int iNumSol = 1;
    boolean bPrintSol = TRUE;
    unsigned long lMaxTime = 0;
    float fTmp;
    unsigned long lMaxBt = 1000000;
    int iLearnOrder = 3;
    int iInputFileArg = 0;
    int iMaxToTry = 10;
    float fFudgeA = 0.8;
    float fFudgeB = 0.9;
    int a=0;
    while (++a < iArgc_){
	if (strcmp(aArgv_[a],"-seed")==0)
	    sscanf(aArgv_[++a], "%ld", &lSeed);
	else if (strcmp(aArgv_[a],"-maxsec")==0){
	    sscanf(aArgv_[++a], "%f", &fTmp);
	    lMaxTime = (unsigned long)(lClkTck * fTmp);
	}
	else if (strcmp(aArgv_[a],"-maxbt")==0)
	    sscanf(aArgv_[++a], "%ld", &lMaxBt);
	else if (strcmp(aArgv_[a],"-numsol")==0)
	    sscanf(aArgv_[++a], "%i", &iNumSol);
	else if (strcmp(aArgv_[a],"-trials")==0)
	    sscanf(aArgv_[++a], "%i", &iNumTrials);
	else if (strcmp(aArgv_[a],"-fudge")==0){
	    sscanf(aArgv_[++a], "%i", &iMaxToTry);
	    sscanf(aArgv_[++a], "%f", &fFudgeA);
	    sscanf(aArgv_[++a], "%f", &fFudgeB);
	}
	else if (strcmp(aArgv_[a],"-nosol")==0)
	    bPrintSol = FALSE;
	else if (strcmp(aArgv_[a],"-norelevance")==0)
	    RelevanceFlag = TRUE;
	else if (strcmp(aArgv_[a],"-dump")==0)
	    DumpFlag = TRUE;
	else
	    iInputFileArg = a;
    }
    if (iInputFileArg <= 0) main(0, aArgv_);
    
    cout << "rel_sat_r 1.0" << endl <<
	"instance: " << aArgv_[iInputFileArg] << endl <<
	"seed: " << lSeed << endl <<
	"cutoff seconds: "<< ((double)lMaxTime)/lClkTck << endl <<
	"cutoff backtracks: "<< lMaxBt << endl <<
	"fudge: " << iMaxToTry << " " << fFudgeA << " " << fFudgeB << endl <<
	"trials: " << iNumTrials << endl <<
	"learning: " << iLearnOrder << endl <<
	"number solutions: " << iNumSol << endl ;

    cout << "dump flag is " << DumpFlag << endl;

    if (DumpFlag){
	char FileP[512];
	char FileT[512];
	strcpy(FileP, aArgv_[iInputFileArg]);
	if (strlen(FileP)>4 && strcmp(&FileP[strlen(FileP)-4], ".cnf")==0)
	    FileP[strlen(FileP)-4] = 0;
	strcpy(FileT, FileP);
	strcat(FileP, "-perm.f");
	strcat(FileT, "-temp.f");
	PermanentFS = new ofstream(FileP);
	TemporaryFS = new ofstream(FileT);
    }

    Random::vInitRandom(lSeed);    

    boolean bFailed;
    int iSuccessCount = 0;
    boolean FinalResult;

    VariableID * pCnfArray;
    VariableID vNvars;
    int iNclauses;
    LoadCnf(aArgv_[iInputFileArg], &vNvars, &iNclauses, &pCnfArray);
    cout << "number variables: " << vNvars << endl;
    cout << "number clauses: " << iNclauses << endl << endl;
    cout << "[trial] solver file result time backtracks branches assigns avg_choices seed\n" << endl;

    SATSolver * xSATSolverPtr;

    for (int i=1; i<=iNumTrials; i++) {
	cout << "[" << i << "] Start trial" << endl;
	SATInstance::xSATInstance.vDestroy();
	/* if (!SATInstance::xSATInstance.bReadDimacs(aArgv_[iInputFileArg]))
	    exit(3); */

	if (!SATInstance::xSATInstance.bReadDimacs(vNvars, iNclauses, pCnfArray))
	    exit(3);

	SATSolver xSATSolver;
	xSATSolverPtr = &xSATSolver;
	boolean bSolution = xSATSolver.bFindSolution(iLearnOrder, bFailed, lMaxTime, lMaxBt, iMaxToTry, fFudgeA, fFudgeB);

	if (! bFailed) { 
	    iSuccessCount++;
	    FinalResult = bSolution;
	    cout << "[" << i << "] SOLVED - instance is " << ( bSolution ? "satisfiable" : 
							       "unsatisfiable") << endl;
	}
	cout << "[" << i << "] rel_sat_r " << aArgv_[iInputFileArg] << 
	    ( bFailed ? " timeout " : ( bSolution ? " SAT " : " UNSAT ")) <<
	    ((double)xSATSolver.lElapsedTime()) / lClkTck << " " <<
	    xSATSolver.lBacktracks() << " " <<
	    xSATSolver.lBranchPointsExplored()  << " " <<
	    xSATSolver.lVariablesLabelled() << " " <<
	    xSATSolver.dAvgChoiceSize() << " " <<
	    lSeed << endl;
    }
    if (iSuccessCount > 0)
	cout << "The instance is " << (FinalResult ? "satisfiable" : "unsatisfiable") << endl;
    cout << endl << "Success rate: " << ((double)iSuccessCount)  / iNumTrials << endl;
    if (bPrintSol && iSuccessCount > 0 && FinalResult){
	(* xSATSolverPtr).vPrintSoln();
    }
}


